from .scVital import makeScVital, scVitalModel, loadModel
#from .lss import calcPairsLSS, calcLSS, calcAUC, calcClustDist, calcTotalDist
#from .merging import mergeAdatas

from . import lss as lss
from . import merging as mg

__all__ = ['makeScVital', 'scVitalModel', "loadModel", "lss", "mg"]


import pytest
import pandas as pd
import numpy as np
from anndata import AnnData
from scipy.sparse import csr_matrix
import scanpy as sc

from scVital.merging import mergeAdatas

@pytest.fixture
def adata_human():
    obs = pd.DataFrame(index=[f"cell_{i}" for i in range(10)])
    var = pd.DataFrame(index=[f"GENE_{i}" for i in range(1000)])
    X = np.random.rand(10, 1000)
    return AnnData(X=X, obs=obs, var=var)

@pytest.fixture
def adata_mouse():
    obs = pd.DataFrame(index=[f"cell_{i}" for i in range(10)])
    var = pd.DataFrame(index=[f"Gene_{i}" for i in range(1000)])
    X = np.random.rand(10, 1000)
    return AnnData(X=X, obs=obs, var=var)

@pytest.fixture
def homology():
    data = {
        'human': [f"GENE_{i}" for i in range(1000)],
        'mouse': [f"Gene_{i}" for i in range(1000)]
    }
    return pd.DataFrame(data)


def test_merge_adatas(adata_human, adata_mouse, homology):
    adatas = [adata_human, adata_mouse]
    species = ["human", "mouse"]
    names = ["inHuman", "inMouse"]
    #data = {
    #    'human': [f"GENE_{i}" for i in range(1000)],
    #    'mouse': [f"Gene_{i}" for i in range(1000)]
    #}
    #homology = pd.DataFrame(data)

    merged_adata = mergeAdatas(adatas, homology=homology, species=species, names=names, label="dataset")
    
    assert merged_adata is not None
    assert merged_adata.obs_names.equals(pd.Index([f"cell_{i}-inHuman" for i in range(10)] + [f"cell_{i}-inMouse" for i in range(10)]))
    assert "species" in merged_adata.obs.columns
    assert merged_adata.obs["species"].unique().tolist() == ["human", "mouse"]
    assert "dataset" in merged_adata.obs.columns
    assert merged_adata.obs["dataset"].unique().tolist() == ["inHuman", "inMouse"]

def testMergeAdatasSameSpec(adata_human):
    adatas = [adata_human, adata_human]

    with pytest.warns(UserWarning, match="No species given, inferring them"):
        merged_adata = mergeAdatas(adatas)
    
    assert merged_adata is not None
    assert "dataset" in merged_adata.obs.columns
    assert merged_adata.obs["dataset"].unique().tolist() == ["adata0human", "adata1human"]
    assert "species" in merged_adata.obs.columns
    assert merged_adata.obs["species"].unique().tolist() == ["human"]

def test_merge_adatas_no_species(adata_human, adata_mouse, homology):
    adatas = [adata_human, adata_mouse]
    #data = {
    #    'human': [f"GENE_{i}" for i in range(1000)],
    #    'mouse': [f"Gene_{i}" for i in range(1000)]
    #}
    #homology = pd.DataFrame(data)

    with pytest.warns(UserWarning, match="No species given, inferring them"):
        merged_adata = mergeAdatas(adatas, homology=homology)
    
    assert merged_adata is not None
    assert "species" in merged_adata.obs.columns

#def test_merge_adatas_no_homology(adata_human, adata_mouse):
#    adatas = [adata_human, adata_mouse]
#    species = ["m1", "m2"]
#    
#    with pytest.warns(UserWarning, match="No homology given, inferring them"):
#        merged_adata = mergeAdatas(adatas, species=species)
#    
#    assert merged_adata is not None
#    assert "species" in merged_adata.obs.columns



#made with copilot

import toml

def get_version():
    with open("pyproject.toml", "r") as f:
        pyproject = toml.load(f)
    return pyproject["project"]["version"]

if __name__ == "__main__":
    print(get_version())


#NOT IMPEMENTED

import os
import time
import numpy as np
import scanpy as sc
import pandas as pd
import seaborn as sns

from matplotlib import cm
import matplotlib
from matplotlib import pyplot as plt
import matplotlib.image as img
import matplotlib.patches as patches


import textwrap


def plotHVG(adata, minMean = 0.05, maxMean = 2.9, minDisp = 0.25, batchKey=None):
	fig, axs = plt.subplots(1, 2, figsize=(10, 5))

	means = adata.var[["means"]][adata.var[["means"]] > np.exp(-14)]#adata.var[["means"]],
	axs[0].hist(np.log(means), bins=100)#, log=True),
	axs[0].axvline(np.log(minMean), color='k', linestyle='dashed', linewidth=1)
	axs[0].axvline(np.log(maxMean), color='k', linestyle='dashed', linewidth=1)
	axs[0].set_title('Gene means counts')
	axs[0].set_xlabel('means')
	axs[0].set_ylabel('counts')

	dispNorm = adata.var[["dispersions_norm"]][adata.var[["dispersions_norm"]] > np.exp(-8)]#adata.var[["means"]],
	axs[1].hist(np.log(dispNorm), bins=100)#, log=True),
	axs[1].axvline(np.log(minDisp), color='k', linestyle='dashed', linewidth=1)
	axs[1].set_title('Gene dispersions counts')
	axs[1].set_xlabel('dispersions')
	axs[1].set_ylabel('counts')

	sc.pp.highly_variable_genes(adata, min_disp=minDisp, min_mean=minMean, max_mean=maxMean, batch_key=batchKey)
	print(sum(adata.var.highly_variable))
	if"highly_variable_intersection" in adata.var.names :
		print(sum(adata.var.highly_variable_intersection))



def vizStats(statFile):
	stats = pd.read_csv(statFile,usecols=[1,2]).T
	vizStatsDF(stats, statFile)

def vizStatsDF(stats, statFile):
	stats = stats.round(3)
	
	# Create a figure and axes
	fig, ax = plt.subplots()

	# Set the number of rows and columns
	num_rows = stats.shape[0]
	num_cols = stats.shape[1]

	# Add a colorbar
	#cbar = plt.colorbar(im)
	for i in range(num_rows):
		for j in range(num_cols):
			value = stats.iloc[i, j]
			if(i==0):
				color = 'green' if value > 0.05 else 'red'
			else:
				color = 'green' if value > 1.25 else 'red'
			#print(f"values {value.round(3)} color {color}")
			rect = plt.Rectangle((j, i), 2, 2, facecolor=color, edgecolor='black', linewidth=1)
			ax.add_patch(rect)

	# Set the tick labels
	ax.set_xticks(np.arange(num_cols) + 0.5)
	ax.set_yticks(np.arange(num_rows) + 0.5)
	ax.set_xticklabels(stats.columns)
	ax.set_yticklabels(stats.index)

	ax.set_xticks(np.arange(num_cols)+1, minor = True)
	ax.set_yticks(np.arange(num_rows)+1, minor = True)

	# Loop over data dimensions and create text annotations
	for i in range(stats.shape[0]):
		for j in range(stats.shape[1]):
			text = ax.text(j+0.5, i+0.5, stats.values[i, j].round(3), ha='center', va='center', color='black', fontsize=12)

	# Set aspect ratio
	ax.set_aspect('equal')

	# Set title and labels
	ax.set_title("Stats")
	plt.xlabel("Clusters")
	plt.ylabel("Metrics")
	fig.savefig(".".join([statFile.split(".csv")[0],"png"]))
	plt.close(fig)

def vizMetrics(metricFile):
	metricDf = pd.read_csv(metricFile, index_col=0).T
	vizMetricsDF(metricDf, metricFile)

def vizMetricsDF(metricDf, metricFile):
	fig, ax = plt.subplots()

	num_rows = metricDf.shape[0]
	num_cols = metricDf.shape[1]

	paletteNum = 50
	rgCmap = sns.diverging_palette(10, 133, n=paletteNum+1)

	for i in range(num_rows):
		for j in range(num_cols):
			value = metricDf.iloc[i, j]
			rect = plt.Rectangle((j, i), 4, 4, edgecolor='black', linewidth=1, facecolor=rgCmap[int(float(value)*paletteNum)])
			ax.add_patch(rect)

	# Set the tick labels
	ax.set_xticks(np.arange(num_cols) + 0.5)
	ax.set_yticks(np.arange(num_rows) + 0.5)
	ax.set_xticklabels(metricDf.columns)
	ax.set_yticklabels(metricDf.index)

	ax.set_xticks(np.arange(num_cols)+1, minor = True)
	ax.set_yticks(np.arange(num_rows)+1, minor = True)

	# Loop over data dimensions and create text annotations
	for i in range(metricDf.shape[0]):
		for j in range(metricDf.shape[1]):
			text = ax.text(j+0.5, i+0.5, metricDf.values[i, j], ha='center', va='center', color='black', fontsize=12)

	# Set aspect ratio
	ax.set_aspect('equal')

	# Set title and labels
	ax.set_title("Metrics")
	plt.xlabel("Clusters")
	plt.ylabel("Metrics")
	fig.savefig(".".join([metricFile.split(".csv")[0],"png"]))
	plt.close(fig)


def plotMetricBarByData(df, label, outDir=None):
	numDatas = df.shape[0]
	dataNames = df.index

	numInteg = df.shape[1]
	integNames = df.columns

	index = np.arange(numDatas)
	width = 1/(numInteg+1)

	bars = np.empty(numInteg, dtype=object)
	fig = plt.figure()
	ax = fig.add_subplot(111)
	for i, integ in enumerate(integNames):
		integVals = np.array(df.loc[:,integ])
		bar = ax.bar(index + width*i, integVals, width)#, color = 'r')
		bars[i] = bar

	ax.set_xlabel("Datasets")
	ax.set_ylabel(label)
	ax.set_title(f"{label}s")
	ax.set_xticks(index + (width*((numInteg-1)/2)), dataNames, rotation=80)
	legend = ax.legend(bars, integNames, loc='center right', bbox_to_anchor=(1.3,0.5))
	fig.savefig(f'{outDir}/{label}_barByData.svg', bbox_inches='tight')
	plt.close(fig)


def plotMetricBarByInteg(df, label, outDir=None):
	numDatas = df.shape[0]
	dataNames = df.index

	numInteg = df.shape[1]
	integNames = df.columns

	index = np.arange(numInteg)
	width = 1/(numDatas+1)

	bars = np.empty(numDatas, dtype=object)
	fig = plt.figure()
	ax = fig.add_subplot(111)
	for i, data in enumerate(dataNames):
		dataVals = np.array(df.loc[data,:])
		bar = ax.bar(index + width*i, dataVals, width)#, color = 'r')
		bars[i] = bar

	ax.set_xlabel("Integration")
	ax.set_ylabel(label)
	ax.set_title(f"{label}s")
	ax.set_xticks(index + (width*((numDatas-1)/2)), integNames,rotation=80)
	legend = ax.legend(bars, dataNames, loc='center right',bbox_to_anchor=(1.3,0.5))
	fig.savefig(f'{outDir}/{label}_barByInteg.svg', bbox_inches='tight')
	plt.close(fig)

def plotScale(scaleDf, outDir=None):
	#scaleDf = (conDict["scale"]/60)

	dataOrder = np.argsort(scaleDf["Size"]).tolist()[::-1]
	datasets = np.array(scaleDf.index.tolist())[dataOrder]
	
	fig, ax = plt.subplots(figsize=(10, 6))
	
	for intAlg in scaleDf.columns[1:].tolist():
		scale = np.array(scaleDf[intAlg])[dataOrder]
		ax.plot(datasets, scale, label=intAlg)
	
	# Adding title and labels
	ax.set_title('Algorithm Performance Comparison')
	ax.set_xlabel('Datasets')
	ax.set_yscale('log')
	ax.set_ylabel('Time  log(minutes)')
	
	# Adding a legend
	ax.legend(loc='center right', bbox_to_anchor=(1.2,0.5))
	
	fig.savefig(f'{outDir}/scale.svg', bbox_inches='tight')
	plt.close(fig)

def df2StackBar(clustBatch, neighborsKey, label, ax):
	clustBatchCount = pd.DataFrame(clustBatch.value_counts(sort=False))
	clusters = np.unique(clustBatch[neighborsKey]).tolist()
	batches = np.unique(clustBatch[label]).tolist()

	counts = pd.DataFrame(np.zeros((len(clusters),len(batches))), index=clusters, columns=batches)

	for clust in clusters:
		for bat in batches:
			try:
				val = clustBatchCount.loc[(clust,bat)].iloc[0]
			except:
				val = 0
			counts.loc[clust,bat] = val

	numClust = len(clusters)#len(adata.obs[neighborsKey].cat.categories)
	rangeClusts = range(0,numClust)

	bott=np.zeros(numClust)
	for bat in counts:
		vals=counts[bat].values
		name=counts[bat].name
		ax.bar(rangeClusts, vals, bottom=bott, label=name)#, color=colorDict[name])
		bott = bott+vals

	ax.set_title(f"# of Cells of each Cluster by {label}") 
	#ax.set_xlabel("Cluster")#neighborsKey
	ax.set_ylabel("# cells")
	ax.legend(loc='center right', bbox_to_anchor=(1.3,0.5))



def vizFracNkBet(clustBatch, neighborsKey, label, kbetScore, outDir=None):
	rangeClusts = range(0,len(kbetScore))

	fig, axs = plt.subplots(2, 1, figsize=(5, 10))

	df2StackBar(clustBatch, neighborsKey, label, axs[0])

	axs[1].bar(rangeClusts, list(kbetScore["kBet"]))
	axs[1].axhline(0.05, color="black", linestyle="--")
	#axs[1].set_ylim(top=1)
	axs[1].set_title("kBet")
	axs[1].set_xlabel("Cluster")#neighborsKey
	axs[1].set_ylabel("kBet Score")

	#fig.legend()
	if(not outDir == None):
		plt.savefig(f'{outDir}/{neighborsKey}_ClustByBatchWkBet.svg')



def vizAllStats(allStats, name="", outDir=None):
	matrix = allStats.values
	xLabels = allStats.index
	yLabels = allStats.columns

	# Create a custom colormap for the first three rows
	RdYlGnCmap = matplotlib.colormaps['RdYlGn']

	# Create a figure and axis
	fig, ax = plt.subplots(figsize=(3.5, 1.5))
	
	# Plot rectangles for each value
	for i in range(matrix.shape[0]):
		for j in range(matrix.shape[1]):
			value = matrix[i, j]
			color = RdYlGnCmap(getCmapValue(value,matrix[i, :]))
			rect = Rectangle((j - 0.5, i - 0.5), 1, 1, facecolor=color, edgecolor='black')
			ax.add_patch(rect)
			textColor='black'
			if(0.212*color[0]+0.7152*color[1]+0.0722*color[3]<0.6):
				textColor='white'
			ax.text(j, i, f"{value:.2f}", ha='center', va='center', color=textColor)
	
	# Set axis labels
	ax.set_xticks(np.arange(-0.5,matrix.shape[1],0.5), [yLabels[i//2] if i%2 ==1 else "" for i,_ in enumerate(np.arange(-0.5,matrix.shape[1],0.5))])
	ax.set_yticks(np.arange(-0.5,matrix.shape[0],0.5), [xLabels[i//2] if i%2 ==1 else "" for i,_ in enumerate(np.arange(-0.5,matrix.shape[0],0.5))])
	ax.set_xticklabels(ax.get_xticklabels(),rotation=70)
	
	# Set plot title
	ax.set_title("Integration Statistics")
	#fig.colorbar(np.arange(0,1,0.2),ax=ax)
	
	# Show the plot
	plt.show()
	
	fig.savefig(f"{outDir}/allStats_{name}.svg", format="svg")


def plotOneUmap(title,x, y, c, edgecolors, name="", linewidths=0.2, s=2, alpha=0.5, outDir=None):
	fig, ax = plt.subplots(1, figsize= (4,4), dpi=300)

	ax.scatter(x, y, c=c, edgecolors=edgecolors, linewidths=linewidths, s=s, alpha=alpha)
	ax.set_xlabel('UMAP 0')
	ax.set_ylabel('UMAP 1')
	ax.set_xticks([])
	ax.set_yticks([])
	ax.set_title(title)

	plt.tight_layout()
	plt.show()

	if("/" in title):
		title = "".join(title.split("/"))
	fig.savefig(f"{outDir}/umap_{title}_{name}.png", format="png")


def makeLegend(ctVals, btVals, cellTypeColorDict, batchColorDict, name="", outDir=None):
	
	ctColLab = cellTypeColorDict.values()
	btColLab = batchColorDict.values()

	ctLegendEle = [Line2D([0], [0], color=ctc, marker='o', lw=0, label=ctLabel) for ctLabel,ctc in ctColLab]
	spaceLegEle = [Line2D([0], [0], marker='o', lw=0, color='white', markeredgecolor='white', label="")]
	btLegendEle = [Line2D([0], [0], color=btc, marker='o', lw=0, markeredgecolor='black', label=btLabel) for btLabel,btc in btColLab]

	ctColors = np.array([cellTypeColorDict[ct][1] for ct in ctVals])
	btColors = np.array([batchColorDict[bt][1] for bt in btVals])

	legendEle = btLegendEle + spaceLegEle + ctLegendEle

	fig, ax = plt.subplots(1, dpi=300)
	plt.legend(handles=legendEle)
	plt.axis('off')
	plt.tight_layout()
	plt.show()
	
	fig.savefig(f"{outDir}/legend_{name}.svg", format="svg")
	
	return(legendEle, ctColors, btColors)

def plotInteg(inUmaps, titles, ctColors, btColors, shuff, outDir=None):
	for i, iUmap in enumerate(inUmaps):
		plotOneUmap(titles[i], x=iUmap[shuff, 0],y=iUmap[shuff, 1], c=ctColors[shuff], edgecolors=btColors[shuff])

def plotCbar(title,name, norm, cmap):
	fig, ax = plt.subplots(1, dpi=300)
	fig.colorbar(cm.ScalarMappable(norm=norm, cmap=cmap), ax=ax)
	plt.axis('off')
	plt.tight_layout()
	plt.show()
	title = "".join(title.split("/"))
	fig.savefig(f"{outDir}/cbarleg_{title}_{name}.svg", format="svg")




#made with the help of copilot

import pytest
import numpy as np
import pandas as pd
import anndata as an
from scipy.sparse import csr_matrix

import torch
import warnings

import scVital as scVt

@pytest.fixture
def setupData():
    counts = csr_matrix(np.random.poisson(1, size=(100, 2000)), dtype=np.float32)

    adata = an.AnnData(counts)
    adata.obs_names = [f"Cell_{i:d}" for i in range(adata.n_obs)]
    adata.var_names = [f"Gene_{i:d}" for i in range(adata.n_vars)]
    bt = np.random.choice(["b1", "b2"], size=(adata.n_obs,))
    adata.obs["batch"] = pd.Categorical(bt)  # Categoricals are preferred for efficiency

    ct = np.random.choice(["ct1", "ct2"], size=(adata.n_obs,))
    adata.obs["cellType"] = pd.Categorical(ct)  # Categoricals are preferred for efficiency

    return {
        'adata': adata,
        'batchLabel': 'batch',
        'miniBatchSize': 512,
        'numEpoch': 64,
        'learningRate': 3e-1,
        'hid1': 1024,
        'hid2': 128,
        'latentSize': 12,
        'discHid': 6,
        'reconCoef': 2e0,
        'klCoef': 1e-1,
        'discCoef': 1e0,
        'discIter': 5,
        'earlyStop': 1e-2,
        'train': False,
        'seed': 18,
        'verbose': True
    }

def test_invalid_adata(setupData):
    with pytest.raises(ValueError):
        scVt.makeScVital(None, setupData['batchLabel'])

def test_invalid_batchLabel(setupData):
    with pytest.raises(ValueError):
        scVt.makeScVital(setupData['adata'], 123)

def test_invalid_miniBatchSize(setupData):
    with pytest.raises(ValueError):
        scVt.makeScVital(setupData['adata'], setupData['batchLabel'], miniBatchSize=-1)

def test_high_learningRate_warning(setupData):
    with pytest.warns(UserWarning):
        scVt.makeScVital(setupData['adata'], setupData['batchLabel'], learningRate=2e0)

def test_valid_makeScVital(setupData):
    model = scVt.makeScVital(**setupData)
    assert isinstance(model, scVt.scVitalModel)

def test_scVital_initialization(setupData):
    model = scVt.scVitalModel(
        setupData['adata'], setupData['batchLabel'], setupData['miniBatchSize'], setupData['numEpoch'], setupData['learningRate'],
        setupData['hid1'], setupData['hid2'], setupData['latentSize'], setupData['discHid'], 
        setupData['reconCoef'], setupData['klCoef'], setupData['discCoef'], setupData['discIter'], 
        setupData['earlyStop'], setupData['seed'], setupData['verbose']
    )
    assert csr_matrix.sum(model.getAdata().X) == csr_matrix.sum(setupData['adata'].X)
    assert model.getBatchLabel() == setupData['batchLabel']
    assert model.getMiniBatchSize() == setupData['miniBatchSize']
    assert model.getNumEpoch() == setupData['numEpoch']
    assert model.getLearningRate() == setupData['learningRate']
    assert model.getLayerDims() == [len(setupData['adata'].var_names), setupData['hid1'], setupData['hid2'], setupData['latentSize']]
    assert model.getLatentSize() == setupData['latentSize']
    assert model.getDiscDims() == [setupData['latentSize'], setupData['discHid'], 2] 
    assert model.getReconCoef() == (len(setupData['adata'])**0.5)*setupData['reconCoef']
    assert model.getKlCoef() == setupData['klCoef']
    assert model.getDiscCoef() == setupData['discCoef']
    assert model.getDiscIter() == setupData['discIter']
    assert model.getEarlyStop() == setupData['earlyStop']
    #assert model.seed == setupData['seed']
    #assert model.verbose == setupData['verbose']



#code to merge adata boject for ease of use to run in scVital

import pandas as pd
import numpy as np
import scanpy as sc
import anndata as ad
import warnings

from scipy.sparse import csr_matrix


def mergeAdatas(adatas, homology=None, species=None, names=None, label="dataset"):
    """
    Merge multiple AnnData objects, inferring species and homology if not provided.

    Parameters
    ----------
    adatas : list of AnnData
        List of AnnData objects to be merged.
    homology : pandas.DataFrame, optional
        DataFrame containing homology information between species.
    species : list of str, optional
        List of species corresponding to each AnnData object.
    names : list of str, optional
        List of names for each AnnData object.
    label : str, optional
        Label for the concatenated AnnData object. Default is "dataset".

    Returns
    -------
    AnnData
        A concatenated AnnData object containing all input AnnData objects.
    """
    if species is None:
        warnings.warn("No species given, inferring them")
        species = []
        for adata in adatas:
            if "species" in adata.obs.columns:
                species.append(adata.obs["species"].unique()[0])
            else:
                if sum([gene.upper() == gene for gene in adata.var_names.values]) > 0.8 * len(adata.var_names):
                    species.append("human")
                    adata.obs["species"] = pd.Categorical(["human"] * len(adata.obs_names))
                if sum([gene[0].upper() + gene[1:].lower() == gene for gene in adata.var_names.values]) > 0.8 * len(adata.var_names):
                    species.append("mouse")
                    adata.obs["species"] = pd.Categorical(["mouse"] * len(adata.obs_names))
        warnings.warn(f"Species inferred: {species}")
    allGenes = set()
    for adata in adatas:
        if "species" not in adata.obs.columns:
            if sum([gene.upper() == gene for gene in adata.var_names.values]) > 0.8 * len(adata.var_names):
                adata.obs["species"] = pd.Categorical(["human"] * len(adata.obs_names))
            elif sum([gene[0].upper() + gene[1:].lower() == gene for gene in adata.var_names.values]) > 0.8 * len(adata.var_names):
                adata.obs["species"] = pd.Categorical(["mouse"] * len(adata.obs_names))
            else:
                warnings.warn("Not mouse or human and not given")
                adata.obs["species"] = pd.Categorical(["unk"] * len(adata.obs_names))
        allGenes.update(list(adata.var_names.values))

    if homology is None and len(set(species)) > 1:
        warnings.warn("No homology given, inferring them")
        if species is None:
            raise ValueError("Species information is required to infer homology")
        species = [sp.lower() for sp in species]
        if "mouse" in species and "human" in species:
            homology = pd.read_csv("../data/homology/MouseHumanHomology.csv")
        else:
            raise ValueError("Only mouse and human homology is known by default")
    elif len(set(species)) == 1:
        homology = pd.DataFrame({species[0]:list(allGenes)})
    geneSpecDict = []
    for i, adata in enumerate(adatas):
        try:
            if csr_matrix.max(adata.X) - csr_matrix.min(adata.X) > 20:
                warnings.warn("Data not log normalized, calculating now")
                sc.pp.normalize_total(adata, target_sum=1e4)
                sc.pp.log1p(adata)
                #warnings.warn(f"Post-normalization range: {csr_matrix.max(adata.X) - csr_matrix.min(adata.X)}")
        except:
            if np.max(adata.X) - np.min(adata.X) > 20:
                warnings.warn("Data not log normalized, calculating now")
                sc.pp.normalize_total(adata, target_sum=1e4)
                sc.pp.log1p(adata)
                #warnings.warn(f"Post-normalization range: {np.max(adata.X) - np.min(adata.X)}")
        geneSpecDict.append({"genes": np.array(adata.var_names.copy()), "species": species[i]})

    geneSpecDict = getOverlapGenesMulti(homology, geneSpecDict)
    for i, genesDict in enumerate(geneSpecDict):
        adatas[i].var_names = genesDict["genes"]

    if names is None:
        warnings.warn("No names given, inferring them")
        names = [f"adata{i}{specie}" for i, specie in enumerate(species)]
        warnings.warn(f"Names inferred: {names}")

    mAdata = ad.concat(adatas, join="outer", label=label, keys=[name for name in names], fill_value=0, index_unique="-")
    return mAdata

def getOverlapGenesMulti(homology, geneSpecDict, sep="/"):
    """
    Get overlapping genes across multiple species using homology information.

    Parameters
    ----------
    homology : pandas.DataFrame
        DataFrame containing homology information between species.
    geneSpecDict : list of dict
        List of dictionaries containing genes and species information.
    sep : str, optional
        Separator used to join gene names. Default is "/".

    Returns
    -------
    list of dict
        Updated list of dictionaries with overlapping genes.
    """
    for data in geneSpecDict:
        genes = data["genes"]
        species = data["species"]
        gene_121 = np.array(homology.loc[:, species.lower()])
        for i, gene in enumerate(genes):
            allGenes = gene
            if gene in gene_121:
                allGenes = sep.join(homology.loc[np.where(gene_121 == gene)[0][0], :])
            genes[i] = allGenes
    return geneSpecDict



#NOT IMPEMENTED

import os
import time
import numpy as np
import scanpy as sc
import pandas as pd
import seaborn as sns

from matplotlib import cm
import matplotlib
from matplotlib import pyplot as plt
import matplotlib.image as img
import matplotlib.patches as patches

from sklearn import metrics
from sklearn.neighbors import NearestNeighbors, KNeighborsRegressor
from sklearn.metrics import RocCurveDisplay, roc_curve
from scipy.spatial.distance import squareform
from scipy.cluster.hierarchy import linkage
from scipy.cluster import hierarchy 
from scipy.stats import power_divergence
from scipy.sparse import isspmatrix
from scipy.spatial import distance

#n_pcs -> last layer dims
def kBet(adata,  batchID= "batch", numNeighbors = 50, interations = 20, cellsPerIter = 50, rep="X_pcAE",n_pcs=8):
	uniqueBatch = np.array(adata.obs[batchID].cat.categories)
	numBatch = len(uniqueBatch)
	batchDict = dict(zip(uniqueBatch,range(0,numBatch)))
	batch = np.array([batchDict[x] for x in adata.obs[batchID]])
	uniqueBatch = np.unique(batch)

	totalPropCells = [np.mean(batch == uniqueBatch[i]) for i in range(numBatch)]
	totalProp = np.multiply([totalPropCells]*cellsPerIter,numNeighbors)#np.round()
	totalPvalue = np.zeros(interations*cellsPerIter)
	
	if any([prop > 0.90 for prop in totalPropCells]):
		return(-0.01)

	#TODO change neighbor matrix to latent name and neighbor matrix
	
	sc.pp.neighbors(adata, n_neighbors=numNeighbors+1, n_pcs=n_pcs, use_rep=rep, key_added=f"leiden{numNeighbors+1}")
	neighborMatrix = adata.obsp[f"leiden{numNeighbors+1}_distances"] 
	#neighborMatrix = adata.obsp[f"{rep.split("_")[1]}_distances"] 

	for i in range(interations):
		indices = np.random.choice(len(adata), size=cellsPerIter)
		propCell = [np.bincount(batch[neighborMatrix[indices].nonzero()[1][neighborMatrix[indices].nonzero()[0] == j]], minlength=numBatch) for j in range(cellsPerIter)]
		chiSqOutPvalue = 0
		try:
			chiSqOut = power_divergence(propCell, f_exp=totalProp, axis=1, lambda_=1)
			chiSqOutPvalue = chiSqOut.pvalue
		except:
			chiSqOutPvalue = -0.03
		totalPvalue[i*cellsPerIter:i*cellsPerIter+cellsPerIter] = chiSqOutPvalue
	return(np.median(totalPvalue))

def inverSampson(countMatrix):
	n = sum(countMatrix)
	typeSum = 0
	for i in countMatrix:
		if i > 0:
			typeSum = typeSum + (i*(i-1))
	return((n*(n-1))/typeSum)

#n_pcs -> out layer dims
def getClustMetrics(adata, clusterID, batchID= "batch", numNeighbors = 50, interations = 20, cellsPerIter = 40, rep="X_pcAE",n_pcs=8):
	clusters = list(adata.obs[clusterID].cat.categories)
	metricNames = ["kBet"] #,"iLSI"
	stats = pd.DataFrame(np.zeros((len(clusters),len(metricNames))), index=clusters, columns=metricNames)
	for i, ctype in enumerate(clusters):
		adataClust = adata[[cellType == ctype for cellType in adata.obs[clusterID]],:]
		#pdb.set_trace()
		stats.loc[ctype,"kBet"] = kBet(adataClust, batchID=batchID, numNeighbors=numNeighbors, 
			interations=interations, cellsPerIter=cellsPerIter, rep=rep, n_pcs=n_pcs)
		#stats.loc[ctype,"iLSI"] = inverSampson(adataClust.obs[batchID].value_counts())

	stats["SortCluster"] = stats.index.astype(int)
	stats = stats.sort_values("SortCluster")
	return(stats)
		#print(f"kBet p-value: {np.around(kBetScore, decimals=2)}")
		#print(f"invSamp index: {np.around(isIndex, decimals=2)}")


def testClustAndStats(adata, umapKey, neighborsKey, pcaRep, 
					  cellLabel, batchLabel, res, numOutLayer, outClustStatDir,
					  nNeighborsUMAP=25, nNeighborsKbet=45, inters=25, cellsPerIter=30,
					  save=True):
	sc.set_figure_params(scanpy=True, dpi_save=150, fontsize=24, format='svg', frameon=False,transparent=True)
	
	batchUmapFilename = f"_{batchLabel}_{neighborsKey}"
	neighUmapFilename = f"_{neighborsKey}"
	trueUmapFilename = f"_{cellLabel}_{neighborsKey}"
	if(not save):
		batchUmapFilename, neighUmapFilename, trueUmapFilename = False, False, False

	#adata.uns[f'{batchLabel}_colors'] = ['#FF7F50','#76EEC6']
	if (neighborsKey=="BBKNN"):
		import bbknn
		startTrain = time.process_time() 
		sc.external.pp.bbknn(adata, batch_key=batchLabel, use_rep="X_pca")#, neighbors_within_batch=4, n_pcs=numOutLayer)#, key_added=umapKey)
		endTrain = time.process_time()
		scale = pd.DataFrame(np.array([[adata.X.size, endTrain-startTrain]]), columns=["Size", "BBKNN"])
		scale.to_csv(f"{outClustStatDir}scale_BBKNN.csv")
	else:
		sc.pp.neighbors(adata, n_neighbors=nNeighborsUMAP, n_pcs=numOutLayer, use_rep=pcaRep, key_added=umapKey)

	sc.pp.pca(adata, svd_solver="arpack")
	sc.tl.umap(adata, neighbors_key = umapKey)

	if(cellLabel == "None"):
		cellLabel = np.nan
		trueUmapFilename = False

	#res = 0.5
	#maxAri = 0
	
	if(not pd.isna(cellLabel)):
		#adata = auc.checkPairsAddSimple(adata, batchLabel, cellLabel)
		cellLabel = "overlapLabel"
		#print(" \t YES TRUE CLUSTERING")
		maxARI = 0
		maxRes = 0
		#sc.pp.neighbors(adata, n_neighbors=nNeighborsUMAP, n_pcs=numOutLayer, use_rep=pcaRep, key_added=umapKey)
		
		for i, res in enumerate(np.arange(0.05, 0.8, 0.05)):
			if (neighborsKey=="BBKNN"):
				res=res*0.1
			
			try:
				sc.tl.leiden(adata, resolution=res, key_added = neighborsKey, neighbors_key = umapKey)#, flavor="igraph", n_iterations=2,  directed=False)
			except:
				sc.pp.neighbors(adata, n_neighbors=nNeighborsUMAP, n_pcs=numOutLayer, use_rep=pcaRep, key_added=umapKey)
				sc.tl.leiden(adata, resolution=res, key_added = neighborsKey, neighbors_key = umapKey)#, flavor="igraph", n_iterations=2,  directed=False)

			newARI = metrics.adjusted_rand_score(adata.obs[cellLabel], adata.obs[neighborsKey])

			if (newARI > maxARI):
				maxARI = newARI
				maxRes = res
		
			if(newARI==1 or maxARI > (newARI+0.1)):
				break
		
		sc.pp.neighbors(adata, n_neighbors=nNeighborsUMAP, n_pcs=numOutLayer, use_rep=pcaRep, key_added=umapKey)
		sc.tl.leiden(adata, resolution=maxRes, key_added = neighborsKey, neighbors_key = umapKey)
		
		#sc.pl.umap(adata, color = [cellLabel], ncols = 1, show=False, save=trueUmapFilename, legend_fontsize="xx-small")#palette="tab10",
		metricDF = getClusterMetricDF(labels_true = adata.obs[cellLabel], labels = adata.obs[neighborsKey], neighborsKey=neighborsKey)
		metricDF.to_csv(f"{outClustStatDir}metrics_{neighborsKey}.csv")
	else:
		sc.tl.leiden(adata, resolution = res, key_added = neighborsKey, neighbors_key = umapKey)
		
	#pdb.set_trace()
	#sc.pl.umap(adata, color = [batchLabel], ncols = 1, show=False, save=batchUmapFilename, legend_fontsize="xx-small") #, palette=colorDict,
	#sc.pl.umap(adata, color = [neighborsKey], ncols = 1, show=False, save=neighUmapFilename, legend_fontsize="xx-small")#palette="Set2",

	sc.pl.umap(adata, color = [neighborsKey,batchLabel,cellLabel], ncols = 3, show=False, save=neighUmapFilename, legend_fontsize="xx-small")#palette="Set2",
	sc.pl.pca(adata, color = [neighborsKey,batchLabel,cellLabel], ncols = 3, show=False, save=neighUmapFilename, legend_fontsize="xx-small")
	
	#batchColorDict = dict(zip(adata.obs[batchLabel].unique(), adata.uns[f'{batchLabel}_colors']))
	#cellTColorDict = dict(zip(adata.obs[cellLabel].unique(), adata.uns[f'{cellLabel}_colors']))
	#bCmap = [batchColorDict[x] for x in adata.obs[batchLabel]]
	#cCmap = [cellTColorDict[x] for x in adata.obs[cellLabel]]
	#latentViz = sns.clustermap(adata.obsm[pcaRep][:,0:numOutLayer], cmap="bwr",row_colors=[bCmap,cCmap],col_cluster=False).figure
	#latentViz.savefig(f"{outClustStatDir}latentHeatmap_{neighborsKey}.png")

	aeStats = getClustMetrics(adata, neighborsKey, batchID = batchLabel,
								numNeighbors = nNeighborsKbet, 
								interations = inters, 
								cellsPerIter = cellsPerIter, 
								rep = pcaRep,
								n_pcs = numOutLayer)
	aeStats.to_csv(f"{outClustStatDir}stats_{neighborsKey}.csv")

	clustBatch = adata.obs[[neighborsKey,batchLabel]]
	vizFracNkBet(clustBatch=clustBatch, neighborsKey=neighborsKey, label=batchLabel, kbetScore=aeStats, outDir=outClustStatDir)


	
#	scale = pd.DataFrame(np.array([[adata.X.size, ]]), columns=["Size", "Time"])
#	scale.to_csv(f"{outClustStatDir}scale_{neighborsKey}.csv")


def getClusterMetricDF(labels_true, labels, neighborsKey):
	metricDict = {
				#   'Homogeneity': metrics.homogeneity_score,
				#	'Completeness': metrics.completeness_score,
				#	'V-measure': metrics.v_measure_score,
					'FM':metrics.fowlkes_mallows_score,
					'ARI': metrics.adjusted_rand_score,
				#	'Adjusted Mutual Information': metrics.adjusted_mutual_info_score
				}
	metricOut = [f"{metricDict[metricF](labels_true, labels):0.3}" for metricF in metricDict]
	metricDF = pd.DataFrame(metricOut, columns=[neighborsKey], index=list(metricDict.keys()))
	return(metricDF)


def heirSimi(adata, latent, cellLabel, allCellTypes, clustMet = "cosine"):
	latLen = range(adata.obsm[latent].shape[1])

	clustDist2 = pd.DataFrame(np.zeros((len(allCellTypes),len(latLen))), columns=latLen, index=allCellTypes)
	for ct in clustDist2.index:
		ctMean = np.mean(adata[adata.obs[cellLabel]==ct].obsm[latent], axis=0)
		clustDist2.loc[ct] = ctMean
	#clustDist2

	sns.clustermap(clustDist2, metric = clustMet)


def getAllStats(dirName,datasetName,paramName):
    #allStats = pd.DataFrame(np.zeros((4,6)),index=["ARI","FM","nKbet","LSS"],columns=["scVital","normal","BBKNN","Harmony","scVI","scDREAMER"])
    allStats = pd.DataFrame(np.zeros((4,6)),index=["Runtime\n(min)  ","ARI","FM","LSS"],columns=["scVital","normal","BBKNN","Harmony","scVI","scDREAMER"])

    lss = pd.read_csv(f"{dirName}/{datasetName}/{paramName}/figures/fullAucScores.csv",index_col=0)
    for lat in allStats.columns:
        #for stat in allStats.index:
        met = pd.read_csv(f"{dirName}/{datasetName}/{paramName}/figures/metrics_{lat}.csv",index_col=0)
        allStats.loc["ARI",lat] = met.loc["ARI",lat]    
        allStats.loc["FM",lat] = met.loc["FM",lat]
 
        try:
            scale = pd.read_csv(f"{dirName}/{datasetName}/{paramName}/figures/scale_{lat}.csv",index_col=0)
        except:
            scale = pd.read_csv(f"{dirName}/{datasetName}/{paramName}/figures/scale_No_Integration.csv",index_col=0)
            
        allStats.loc["Runtime\n(min)  ",lat] = (scale[lat][0])/60
        
        #kbet = pd.read_csv(f"{dirName}/{datasetName}/{paramName}/figures/stats_{lat}.csv",index_col=0)
        #allStats.loc["nKbet",lat] = sum(kbet["kBet"]>0.05)/len(kbet["kBet"])
        
        allStats.loc["LSS",lat] = lss.loc[:,f"X_{lat}"].values[0]
    
    return(allStats)

def getCmapValue(value, vals):
    maxVal = np.max(vals)
    if(maxVal > 1):
        return ((-value+maxVal)/(-np.min(vals)+maxVal))
    return value



#    FROM scib-metrics almost exatly
def plot_results_table(df, show = False, save_dir = None):
    """Plot the benchmarking results.
    Parameters
    ----------
    show
        Whether to show the plot.
    save_dir
        The directory to save the plot to. If `None`, the plot is not saved.
    """
    from plottable import ColumnDefinition, Table
    from plottable.cmap import normed_cmap
    from plottable.plots import bar
    embeds = list(df.index[:-1])
    num_embeds = len(embeds)
    cmap_fn = lambda col_data: normed_cmap(col_data, cmap=cm.PRGn, num_stds=2.5)
    _LABELS = "labels"
    _BATCH = "batch"
    _METRIC_TYPE = "Metric Type"
    _AGGREGATE_SCORE = "Aggregate score"
    # Do not want to plot what kind of metric it is
    plot_df = df.drop(_METRIC_TYPE, axis=0)
    # Sort by total score
    plot_df = plot_df.sort_values(by="Total", ascending=False).astype(np.float64)
    plot_df["Method"] = plot_df.index

    # Split columns by metric type, using df as it doesn't have the new method col
    score_cols = df.columns[df.loc[_METRIC_TYPE] == _AGGREGATE_SCORE]
    other_cols = df.columns[df.loc[_METRIC_TYPE] != _AGGREGATE_SCORE]
    column_definitions = [
        ColumnDefinition("Method", width=1.5, textprops={"ha": "left", "weight": "bold"}),
    ]
    # Circles for the metric values
    column_definitions += [
        ColumnDefinition(
            col,
            title=col.replace(" ", "\n", 1),
            width=1,
            textprops={
                "ha": "center",
                "bbox": {"boxstyle": "circle", "pad": 0.25},
            },
            cmap=cmap_fn(plot_df[col]),
            group=df.loc[_METRIC_TYPE, col],
            formatter="{:.2f}",
        )
        for i, col in enumerate(other_cols)
    ]
    # Bars for the aggregate scores
    column_definitions += [
        ColumnDefinition(
            col,
            width=1,
            title=col.replace(" ", "\n", 1),
            plot_fn=bar,
            plot_kw={
                "cmap": cm.YlGnBu,
                "plot_bg_bar": False,
                "annotate": True,
                "height": 0.9,
                "formatter": "{:.2f}",
            },
            group=df.loc[_METRIC_TYPE, col],
            border="left" if i == 0 else None,
        )
        for i, col in enumerate(score_cols)
    ]
    # Allow to manipulate text post-hoc (in illustrator)
    with plt.rc_context({"svg.fonttype": "none"}):
        fig, ax = plt.subplots(figsize=(len(df.columns) * 1.25, 3 + 0.3 * num_embeds))
        tab = Table(
            plot_df,
            cell_kw={
                "linewidth": 0,
                "edgecolor": "k",
            },
            column_definitions=column_definitions,
            ax=ax,
            row_dividers=True,
            footer_divider=True,
            textprops={"fontsize": 10, "ha": "center"},
            row_divider_kw={"linewidth": 1, "linestyle": (0, (1, 5))},
            col_label_divider_kw={"linewidth": 1, "linestyle": "-"},
            column_border_kw={"linewidth": 1, "linestyle": "-"},
            index_col="Method",
        ).autoset_fontcolors(colnames=plot_df.columns)
    if show:
        plt.show()
    if save_dir is not None:
        fig.savefig(os.path.join(save_dir, "scib_results.svg"), facecolor=ax.get_facecolor(), dpi=300)
        plt.close(fig)
    return tab







import pytest
import warnings

import numpy as np
import pandas as pd
import anndata as an
from scipy.sparse import csr_matrix
from sklearn.metrics import roc_auc_score
from scipy.spatial import distance

# Assuming the functions are imported from lss.py
from scVital.lss import calcPairsLSS, calcLSS, calcAUC, calcClustDist, calcTotalDist

# Mock data for testing
@pytest.fixture
def mock_data():
	np.random.seed(1)

	counts = csr_matrix(np.random.poisson(1, size=(100, 2000)), dtype=np.float32)
	latent = 'X_pca'
	batchName = 'batch'
	cellTypeLabel = 'cellType'

	adata = an.AnnData(counts)
	adata.obs_names = [f"Cell_{i:d}" for i in range(adata.n_obs)]
	adata.var_names = [f"Gene_{i:d}" for i in range(adata.n_vars)]

	bt = np.random.choice(["m", "h"], size=(adata.n_obs,))
	adata.obs[batchName] = pd.Categorical(bt)  # Categoricals are preferred for efficiency

	ct = np.random.choice(["ct1", "ct2", "ct3"], size=(adata.n_obs,))
	adata.obs[cellTypeLabel] = pd.Categorical(ct)  # Categoricals are preferred for efficiency

	latentSpace = np.random.normal(loc=0.0, scale=1.0, size=(len(adata),10))
	adata.obsm[latent] = latentSpace
	return adata, latent, batchName, cellTypeLabel

def test_calcPairsLSS(mock_data):
	adata, latent, batchName, cellTypeLabel = mock_data
	clustDist, lssAUC, totalDist, allCellTypes, ctPairs = calcPairsLSS(adata, latent, batchName, cellTypeLabel)
	assert np.sum(clustDist.values) == 31.8386
	assert np.round(lssAUC, decimals=4) == 0.5593
	assert totalDist == 2.3621
	assert allCellTypes == ['h~ct1', 'h~ct2', 'h~ct3', 'm~ct1', 'm~ct2', 'm~ct3']



#!/usr/bin/env python
#run train DL model
#comments and error checking amde with the help of copilot

import os
import sys
import time
import pickle
import warnings

import pandas as pd
import numpy as np
import scanpy as sc
import anndata as an
from scipy.sparse import isspmatrix
from matplotlib import pyplot as plt


import torch
from torch import nn
from torch.utils.data import DataLoader
from torch import optim
import torch.nn.functional as F

#from scVital.autoencoder import EncoderDecoder, Encoder, Decoder
#from scVital.discriminator import Discriminator

def makeScVital(
	adata: an.AnnData,
	batchLabel: str,
	miniBatchSize: int = 512,
	numEpoch: int = 64,
	learningRate: float = 1e-3,
	hid1: int = 1024,
	hid2: int = 128,
	latentSize: int = 12,
	discHid: int = 6,
	reconCoef: float = 2e1,
	klCoef: float = 1e-1,
	discCoef: float = 1e0,
	discIter: int = 5,
	earlyStop: float = 1e-2,
	train: bool=False,
	seed: int = 18,
	verbose: bool = True
) -> 'scVitalModel':
	"""
	Run the scVital model with the specified parameters.

	Parameters:
	adata (an.AnnData): Annotated data matrix.
	batchLabel (str): Label for batch processing.
	miniBatchSize (int, optional): Size of mini-batches for training. Default is 512.
	numEpoch (int, optional): Number of epochs for training. Default is 64.
	learningRate (float, optional): Learning rate for the optimizer. Default is 1e-3.
	hid1 (int, optional): Number of units in the first hidden layer. Default is 1024.
	hid2 (int, optional): Number of units in the second hidden layer. Default is 128.
	latentSize (int, optional): Size of the latent space. Default is 12.
	discHid (int, optional): Number of units in the discriminator hidden layer. Default is 6.
	reconCoef (float, optional): Coefficient for reconstruction loss. Default is 2e1.
	klCoef (float, optional): Coefficient for KL divergence loss. Default is 1e-1.
	discCoef (float, optional): Coefficient for discriminator loss. Default is 1e0.
	discIter (int, optional): Number of iterations for discriminator training. Default is 5.
	earlyStop (float, optional): Delta error to trigger early stopping. Default is 1e-2.
	seed (int, optional): Random seed for reproducibility. Default is 18.
	verbose (bool, optional): Flag for verbosity. Default is True.

	Returns:
	scVital: An instance of the scVital class initialized with the specified parameters.
	
	Raises:
	ValueError: If any of the input parameters are invalid.
	"""
	# Check valid inputs
	if not isinstance(adata, an.AnnData):
		raise ValueError("adata must be an AnnData object")
	if not isinstance(batchLabel, str):
		raise ValueError("batchLabel must be a string")
	if not isinstance(miniBatchSize, int) or miniBatchSize <= 0:
		raise ValueError("miniBatchSize must be a positive integer")
	if not isinstance(numEpoch, int) or numEpoch <= 0:
		raise ValueError("numEpoch must be a positive integer")
	if not isinstance(learningRate, float) or learningRate <= 0:
		raise ValueError("learningRate must be a positive float")
	if not isinstance(hid1, int) or hid1 <= 0:
		raise ValueError("hid1 must be a positive integer")
	if not isinstance(hid2, int) or hid2 <= 0:
		raise ValueError("hid2 must be a positive integer")
	if not isinstance(latentSize, int) or latentSize <= 0:
		raise ValueError("latentSize must be a positive integer")
	if not isinstance(discHid, int) or discHid <= 0:
		raise ValueError("discHid must be a positive integer")
	if not isinstance(reconCoef, float) or reconCoef < 0:
		raise ValueError("reconCoef must be a non-negative float")
	if not isinstance(klCoef, float) or klCoef < 0:
		raise ValueError("klCoef must be a non-negative float")
	if not isinstance(discCoef, float) or discCoef < 0:
		raise ValueError("discCoef must be a non-negative float")
	if not isinstance(discIter, int) or discIter <= 0:
		raise ValueError("discIter must be a positive integer")
	if not isinstance(earlyStop, float) or earlyStop < 0:
		raise ValueError("earlyStop must be a non-negative float")
	if not isinstance(train, bool):
		raise ValueError("train must be a boolean")
	if not isinstance(seed, int) or seed < 0:
		raise ValueError("seed must be a non-negative integer")
	if not isinstance(verbose, bool):
		raise ValueError("verbose must be a boolean")

	# Issue a warning if the learning rate is unusually high
	if learningRate > 1:
		warnings.warn("The learning rate is unusually high and may cause instability.", UserWarning)
	
	# Issue a warning if the adata is very large
	if adata.shape[1] > 4000:
		warnings.warn("The adata object has many genes consider subsetting on highly variable genes", UserWarning)

	# Initialize the scVital model with the provided parameters
	scVitalData = scVitalModel(
		adata, batchLabel, miniBatchSize, numEpoch, learningRate,
		hid1, hid2, latentSize, discHid, reconCoef, klCoef, discCoef,
		discIter, earlyStop, seed, verbose
	)

	# Train is true then train 
	if(train):
		scVitalData.runTrainScVital(self)

	# Return the initialized scVital model
	return scVitalData


class scVitalModel(object):
	def __init__(self, adata, batchLabel, miniBatchSize, numEpoch, learningRate,
				hid1, hid2, latentSize, discHid, reconCoef, klCoef, discCoef, discIter, 
				earlyStop, seed, verbose):
		"""
		Initialize the model with the given parameters.

		Parameters:
		adata (AnnData): Annotated data matrix.
		batchLabel (str): Label for batch processing.
		miniBatchSize (int): Size of mini-batches for training.
		numEpoch (int): Number of epochs for training.
		learningRate (float): Learning rate for the optimizer.
		hid1 (int): Number of units in the first hidden layer.
		hid2 (int): Number of units in the second hidden layer.
		latentSize (int): Size of the latent space.
		discHid (int): Number of units in the discriminator hidden layer.
		reconCoef (float): Coefficient for reconstruction loss.
		klCoef (float): Coefficient for KL divergence loss.
		discCoef (float): Coefficient for discriminator loss.
		discIter (int): Number of iterations for discriminator training.
		earlyStop (float): Delta error to trigger early stopping.
		seed (int): Random seed for reproducibility.
		verbose (bool): Flag for verbosity.
		"""
		self.__adata = adata
		self.__batchLabel = batchLabel
		self.__miniBatchSize = miniBatchSize
		self.__numEpoch = numEpoch
		self.__learningRate = learningRate
		self.__hid1 = hid1
		self.__hid2 = hid2
		self.__latentSize = latentSize
		self.__discHid = discHid
		self.__reconCoef = reconCoef
		self.__klCoef = klCoef
		self.__discCoef = discCoef
		self.__discIter = discIter
		self.__earlyStop = earlyStop
		self.__seed = seed
		self.__verbose = verbose
		self.__lossDict = {"total":[],"recon":[],"trick":[],"klDiv":[],"discr":[]}
		# Set the random seed for reproducibility
		torch.manual_seed(seed)

		# Get training data and labels
		inData, batchSpecLabIndex = self._getTrainLabel(speciesLabel="species")

		# Prepare data by appending labels
		LabeledData, layer1Dim, numSpeices = self._getLabeledData(inData)
		self.__LabeledData = LabeledData
		self.__numSpeices = numSpeices

		# Define layer dimensions
		self.__layerDims = [layer1Dim, hid1, hid2, latentSize]
		self.__inDiscriminatorDims = [latentSize, discHid]

		# Adjust reconstruction coefficient
		self.__reconCoef = self.__reconCoef * (inData.shape[0] ** 0.5)

		# Initialize encoder and decoder
		encoder = Encoder(self.__layerDims, self.__numSpeices)
		decoder = Decoder(self.__layerDims, self.__numSpeices, geneIndexes=batchSpecLabIndex)

		# Initialize autoencoder
		self.__autoencoder = EncoderDecoder(encoder, decoder)
	
		# Initialize discriminator
		self.__discriminator = Discriminator(self.__inDiscriminatorDims, self.__numSpeices)

	def runTrainScVital(self):
		"""
		Train the scVital model, which includes an autoencoder and a discriminator, and store the results.

		This function initializes the encoder, decoder, autoencoder, and discriminator. It sets up the optimizers and learning rate schedulers,
		trains the models, and then stores the trained models and loss information. Finally, it prepares the data for evaluation and stores
		the latent representations and reconstructed data in an AnnData object.

		Attributes:
			self.__layerDims (list): Dimensions of the layers for the encoder and decoder.
			self.__numSpeices (int): Number of species (classes) in the dataset.
			self.__learningRate (float): Learning rate for the optimizers.
			self.__inDiscriminatorDims (list): Dimensions of the layers for the discriminator.
			self.__inLabels (torch.Tensor): Input labels for the data.
			self.__LabeledData (torch.Tensor): Labeled data for training.
			self.__adata (AnnData): AnnData object to store the results.
		"""
		# Initialize autoencoder
		autoencoderOpt = optim.AdamW(params=self.__autoencoder.parameters(), lr=self.__learningRate)
		aeSchedLR = optim.lr_scheduler.CosineAnnealingWarmRestarts(autoencoderOpt, T_0=5, T_mult=2)
		reconstructionLossFunc = torch.nn.MSELoss()

		# Initialize discriminator
		discriminatorOpt = optim.AdamW(params=self.__discriminator.parameters(), lr=self.__learningRate)
		discSchedLR = optim.lr_scheduler.CosineAnnealingWarmRestarts(discriminatorOpt, T_0=5, T_mult=2)
		discriminatorLossFunc = torch.nn.CrossEntropyLoss()

		# Train the model
		self._trainScVital(autoencoderOpt, reconstructionLossFunc, aeSchedLR, discriminatorOpt, discriminatorLossFunc, discSchedLR)

		# Set models to evaluation mode
		self.__autoencoder.eval()
		encoderOut = self.__autoencoder.getEncoder()
		encoderOut.eval()
		self.__discriminator.eval()

		# Prepare one-hot encoded labels
		LabOneHot = torch.reshape(F.one_hot(self.__inLabels.to(torch.int64), num_classes=self.__numSpeices).float(), (self.__LabeledData.shape[0], self.__numSpeices))#inData.shape[0]
		labOneHotInData = torch.cat((self.__LabeledData[:,1:], LabOneHot), axis=1)

		# Get latent representations and reconstructed data
		allEncOut = encoderOut(labOneHotInData)
		bLatent = allEncOut.detach().numpy()
		reconData = self.__autoencoder(labOneHotInData, self.__inLabels, self.__numSpeices).detach().numpy()

		# Store results in AnnData object
		self.__adata.obsm["X_scVital"] = bLatent
		self.__adata.layers["scVitalRecon"] = reconData


	def _trainScVital(self, autoencoderOpt, reconstructionLossFunc, aeSchedLR,
					 discriminatorOpt, discriminatorLossFunc, discSchedLR):
		self.__autoencoder.train()
		self.__discriminator.train()

		ldTrainDataLoader = DataLoader(self.__LabeledData, batch_size=self.__miniBatchSize, shuffle=True)
		numMiniBatch = len(ldTrainDataLoader)

		discEpochLoss = np.full(self.__numEpoch, np.nan)
		reconstEpochLoss = np.full(self.__numEpoch, np.nan)
		trickEpochLoss = np.full(self.__numEpoch, np.nan)
		klDivEpochLoss = np.full(self.__numEpoch, np.nan)
		aeTotalEpochLoss = np.full(self.__numEpoch, np.nan)

		prevTotalLoss = np.inf

		#KL Divergenace Cyclical Annealing
		klDivCoeff = self._klCycle(0, self.__klCoef, self.__numEpoch)

		for iEp in range(self.__numEpoch):

			discTrainLoss = np.full(numMiniBatch, np.nan) 
			reconstTrainLoss = np.full(numMiniBatch, np.nan)
			trickTrainLoss = np.full(numMiniBatch, np.nan)
			klDivTrainLoss = np.full(numMiniBatch, np.nan)
			aeTotalTrainLoss = np.full(numMiniBatch, np.nan)

			for iBatch, ldData in enumerate(ldTrainDataLoader):
				#Load data 
				bRealLabels, bData = ldData[:,0].to(torch.int64), ldData[:,1:].to(torch.float32)
				
				bRealLabOneHot = F.one_hot(bRealLabels, num_classes=self.__numSpeices).float()
				labeledBData = torch.cat((bData, bRealLabOneHot),axis=1)
				labels = np.unique(bRealLabels)
				
				#Train discriminator
				#get mouse and human data in latent space
				encoder = self.__autoencoder.getEncoder()
				#pdb.set_trace()
				for i in range(self.__discIter):
					bLatent = encoder(labeledBData) #bData
					
					#Optimize discriminator
					discriminatorOpt.zero_grad()

					bDiscPred = self.__discriminator(bLatent)
					bDiscRealLoss = discriminatorLossFunc(bDiscPred, bRealLabels)
					bDiscRealLoss.backward()

					discriminatorOpt.step()
					discSchedLR.step(iEp + iBatch / numMiniBatch)

				#Train generator
				autoencoderOpt.zero_grad()
				
				#encode mouse and human data in latent space
				bReconstData = self.__autoencoder(labeledBData, bRealLabels, self.__numSpeices) #bData

				#added
				#split input data and reconstructed data by batch and batch-specific genes
				#calculate reconstruction on on relavent data
				bReconstLoss = torch.tensor(0.0)
				batchSpecLabIndex = self.__autoencoder.getGeneIndex()
				allCells, allGenes = bData.shape

				for i in labels: # for every batch
					sCells = (bRealLabels==i).reshape((allCells,1))	   #Cells realting to label
					#sGenes = torch.tensor(bool(batchSpecLabIndex[i])).reshape((1, allGenes))

					sGenes = torch.tensor(batchSpecLabIndex[i]).reshape((1,allGenes))	 #Genes relating to label
					numCells, numGenes = torch.sum(sCells), torch.sum(sGenes,axis=1)	  #number of cells and genes in of the label
					sMask = sCells * sGenes	   #tensor of the mask to only take the genes realting to labels for the same cells
					bDataMasked = torch.masked_select(bData, sMask).reshape((numCells,numGenes))	  #apply mask to input data
					reconDataMasked = torch.masked_select(bReconstData, sMask).reshape((numCells,numGenes))   #apply mask to reconstructed data
					bReconstLoss += (numCells/allCells)*reconstructionLossFunc(reconDataMasked, bDataMasked)  #calcualte reconstruction with masks and update total
				
				#bReconstLoss = reconstructionLossFunc(bReconstData, bData) 
				
				encoder = self.__autoencoder.getEncoder()
				bLatent = encoder(labeledBData) #bData

				#train discriminator and get preds try and subtract from total Loss
				bDiscPred = self.__discriminator(bLatent)

				#bDiscWrongLoss = discriminatorLossFunc(bDiscPred, bRealLabels)
				bRealLabOneHot = F.one_hot(bRealLabels, num_classes=self.__numSpeices).float()
				reconLatentEven = torch.ones_like(bRealLabOneHot)*(1/self.__numSpeices)
				bDiscTrickLoss = discriminatorLossFunc(bDiscPred, reconLatentEven) #bRealLabels)
				
				#KL Div loss with N(0,1)
				klDivLoss = encoder.klDivLoss

				bRecTrickDiscLoss = self.__reconCoef*bReconstLoss + klDivCoeff[iEp]*klDivLoss + self.__discCoef*bDiscTrickLoss 
				bRecTrickDiscLoss.backward()
				
				#optimize generator
				autoencoderOpt.step()
				aeSchedLR.step(iEp + iBatch / numMiniBatch)

				discTrainLoss[iBatch] 		= bDiscRealLoss.item()
				reconstTrainLoss[iBatch] 	= self.__reconCoef*bReconstLoss.item()
				trickTrainLoss[iBatch]		= self.__discCoef*bDiscTrickLoss.item()
				klDivTrainLoss[iBatch]		= klDivCoeff[iEp]*klDivLoss.item()
				aeTotalTrainLoss[iBatch] 	= bRecTrickDiscLoss.item()

				if (self.__verbose and (iEp % 10 == 0 and iBatch % 8 == 0)):
					print(f"Epoch={iEp}, batch={iBatch}, discr={np.nanmean(discTrainLoss):.4f}, total={np.nanmean(aeTotalTrainLoss):.4f}, recon={np.nanmean(reconstTrainLoss):.4f}, trick={np.nanmean(trickTrainLoss):.4f}, klDiv={np.nanmean(klDivTrainLoss):.4f}")

			discEpochLoss[iEp] 		= np.nanmean(discTrainLoss)
			reconstEpochLoss[iEp] 	= np.nanmean(reconstTrainLoss)
			trickEpochLoss[iEp] 	= np.nanmean(trickTrainLoss)
			klDivEpochLoss[iEp] 	= np.nanmean(klDivTrainLoss)
			aeTotalEpochLoss[iEp] 	= np.nanmean(aeTotalTrainLoss)
			
			#if iEp % 50 == 0:
			#	print(f"Epoch={iEp}, \t	discr={np.nanmean(discEpochLoss):.4f}, total={np.nanmean(aeTotalEpochLoss):.4f}, recon={np.nanmean(reconstEpochLoss):.4f}, trick={np.nanmean(trickEpochLoss):.4f}, klDiv={np.nanmean(klDivEpochLoss):.4f}")

			#Early Stopping
			totalLoss = np.mean(aeTotalEpochLoss[max(iEp-5,0):(iEp+1)])
			deltaLoss = np.abs(prevTotalLoss-totalLoss)
			if (deltaLoss < self.__earlyStop and iEp > 10):
				print(f' epoch:{iEp} delta:{deltaLoss}')
				break
			prevTotalLoss = totalLoss
			
			#aeSchedLR.step()
			#discSchedLR.step()

		self.__lossDict = {"total":aeTotalEpochLoss,
					"recon":reconstEpochLoss,
					"trick":trickEpochLoss,
					"klDiv":klDivEpochLoss,
					"discr":discEpochLoss}
		#return


	def _getLabeledData(self, inData):
		"""
		Concatenate labels and data along the specified axis and return the labeled data,
		the dimension of the first layer, and the number of unique species.

		Parameters:
		inData (torch.Tensor): Input data tensor.
		inLabels (torch.Tensor): Input labels tensor.

		Returns:
		tuple: A tuple containing the labeled data, the dimension of the first layer, and the number of unique species.
		"""
		# Concatenate labels and data along the columns
		LabeledData = torch.cat((self.__inLabels, inData), axis=1)
		
		# Get the dimension of the first layer (number of features in the input data)
		layer1Dim = inData.size()[1]
		
		# Get the number of unique species from the labels
		numSpeices = len(self.__inLabels.unique())
		
		return LabeledData, layer1Dim, numSpeices

	@staticmethod
	def _klCycle(start, stop, n_epoch, n_cycle=4):
		"""
		Generate a KL divergence schedule that cycles between start and stop values over the specified number of epochs.

		Parameters:
		start (float): Starting value of the KL divergence.
		stop (float): Stopping value of the KL divergence.
		n_epoch (int): Total number of epochs.
		n_cycle (int): Number of cycles within the total epochs.

		Returns:
		numpy.ndarray: An array representing the KL divergence schedule.
		"""
		ratio = (n_cycle - 1) / n_cycle
		kl = np.full(n_epoch, stop)
		period = n_epoch / n_cycle
		step = (stop - start) / (period * ratio)  # Linear schedule

		for c in range(n_cycle):
			v, i = start, 0
			while v <= stop and (int(i + c * period) < n_epoch):
				kl[int(i + c * period)] = v
				v += step
				i += 1
		return kl

	def _getAdataX(self):
		"""
		Convert the AnnData object matrix to a dense tensor.

		Parameters:
		adata (AnnData): Annotated data matrix.

		Returns:
		torch.Tensor: Dense tensor representation of the data matrix.
		"""
		adataX = self.__adata.X
		if isspmatrix(adataX):
			adataX = adataX.todense()
		return torch.tensor(adataX)

	def _getTrainLabel(self, speciesLabel="species"):
		"""
		Prepare training data and labels, and generate batch-specific gene indices.

		Parameters:
		speciesLabel (str): Column name for species labels in the AnnData object.

		Returns:
		tuple: A tuple containing the input data, and batch-specific gene indices.
		"""
		# Convert AnnData object matrix to dense tensor
		inData = self._getAdataX()

		# Extract batch labels and create a dictionary mapping unique batches to indices
		batch = np.array(self.__adata.obs[self.__batchLabel])
		uniqueBatch = np.unique(batch)
		numBatch = len(uniqueBatch)
		batchDict = dict(zip(uniqueBatch,range(0,numBatch)))
		batchNum = np.array([batchDict[x] for x in batch])

		# Convert batch numbers to tensor and reshape
		inLabels = torch.tensor(batchNum)
		self.__inLabels = inLabels.view(len(inLabels), 1)

		# Initialize gene type array and batch-species dictionary
		geneType = np.full(len(self.__adata.var_names), "all", dtype=object)
		batchSpeciesDict = {batch: "all" for batch in uniqueBatch}

		# Check if species label exists in the AnnData object
		if np.isin([speciesLabel], self.__adata.obs.columns.values)[0]:
			uniqueSpecies = self.__adata.obs[speciesLabel].cat.categories
			batchSpecies = np.unique(["!".join(x) for x in self.__adata.obs[[self.__batchLabel, speciesLabel]].values])
			batchSpeciesDict = {indBatchSpecies.split("!")[0]: indBatchSpecies.split("!")[1] for indBatchSpecies in batchSpecies}
			
			# Assign gene types based on species
			for i, gene in enumerate(self.__adata.var_names):
				gsplit = gene.split("/")
				if len(gsplit) < 2:
					gInd = np.where(self.__adata.var_names.values == gsplit)[0][0]
					try:
						geneType[i] = self.__adata[(inData[:,gInd]>0).numpy(),:].obs[speciesLabel].unique()[0]
					except:
						geneType[i] = "none"
				if ' ' in gsplit:
					for j, g in enumerate(gsplit):
						if g != " ":
							geneType[i] = uniqueSpecies[j]
			rangeGenes = range(len(geneType))

		# Generate batch-specific gene indices
		batchSpecLabIndex = []
		for batchLab in uniqueBatch:
			speciesSpecGenes = np.logical_or(geneType == "all", geneType == batchSpeciesDict[batchLab])
			batchSpecLabIndex.append(np.array(list(speciesSpecGenes), dtype=bool))

		return inData, batchSpecLabIndex

	def saveDiscrim(self, outDiscFile):
		"""Save the discriminator to file."""
		torch.save(self.__discriminator, outDiscFile)

	def saveAutoenc(self, outVAEFile):
		"""Save the autoencoder to file."""
		torch.save(self.__autoencoderOut, outVAEFile)

	# Getters for the instance variables
	def getAdata(self):
		"""Return the annotated data matrix."""
		return self.__adata

	def getLabeledData(self):
		return self.__LabeledData

	def getBatchLabel(self):
		"""Return the batch label."""
		return self.__batchLabel

	def getMiniBatchSize(self):
		"""Return the mini-batch size."""
		return self.__miniBatchSize

	def getNumEpoch(self):
		"""Return the number of epochs."""
		return self.__numEpoch

	def getLearningRate(self):
		"""Return the learning rate."""
		return self.__learningRate

	def getLayerDims(self):
		"""Return a list of the dimentions of the hidden layers of scVital."""
		return [len(self.__adata.var_names), self.__hid1, self.__hid2, self.__latentSize] 

	def getDiscDims(self):
		"""Return a list of the dimentions of the hidden layers of discriminator scVital."""
		return [self.__latentSize, self.__discHid, self.__numSpeices] 

	def getLatentSize(self):
		"""Return the size of the latent space."""
		return self.__latentSize

	def getReconCoef(self):
		"""Return the coefficient for reconstruction loss."""
		return self.__reconCoef

	def getKlCoef(self):
		"""Return the coefficient for KL divergence loss."""
		return self.__klCoef

	def getDiscCoef(self):
		"""Return the coefficient for discriminator loss."""
		return self.__discCoef

	def getDiscIter(self):
		"""Return the number of iterations for discriminator training."""
		return self.__discIter

	def getEarlyStop(self):
		"""Return the delta error to trigger early stopping."""
		return self.__earlyStop

	def getAutoencoder(self):
		"""Return the Autoencoder."""
		return self.__autoencoder

	def getDisriminator(self):
		"""Return the Discriminator."""
		return self.__discriminator

	def getLossDict(self):
		"""Return the loss dicitonary of training."""
		return self.__lossDict

	def __str__(self):
		return (f"Model Parameters:\n"
			f"adata: {self.__adata}\n"
			f"batchLabel: {self.__batchLabel}\n"
			f"miniBatchSize: {self.__miniBatchSize}\n"
			f"numEpoch: {self.__numEpoch}\n"
			f"learningRate: {self.__learningRate}\n"
			f"reconCoef: {self.__reconCoef}\n"
			f"klCoef: {self.__klCoef}\n"
			f"discCoef: {self.__discCoef}\n"
			f"discIter: {self.__discIter}\n"
			f"earlyStop: {self.__earlyStop}\n"
			f"seed: {self.__seed}\n"
			f"verbose: {self.__verbose}\n"
			f"numSpeices: {self.__numSpeices}\n"
			f"layerDims: {self.__layerDims}\n"
			f"inDiscriminatorDims: {self.__inDiscriminatorDims}")


	def saveModel(self, filename):
		"""
		Save the scVitalModel object to a file.

		Parameters:
		model (scVitalModel): The model object to save.
		filename (str): The name of the file to save the model to.
		"""
		if(filename[-4:] != ".pkl"):
			filename = filename + ".pkl"
		with open(filename, 'wb') as file:
			pickle.dump(self, file)
		print(f"Model saved to {filename}")


	def plotLoss(self, save=False):
		lossFig, lossAxs = plt.subplots(3, 2)
		lossAxs[0, 0].plot(self.__lossDict["recon"]),
		lossAxs[0, 0].set_title('Recon Loss')

		lossAxs[0, 1].plot(self.__lossDict["discr"])
		lossAxs[0, 1].set_title('Disc Loss')

		lossAxs[1, 1].plot(self.__lossDict["trick"])
		lossAxs[1, 1].set_title('Trick Disc Loss')

		lossAxs[1, 0].plot(self.__lossDict["klDiv"])
		lossAxs[1, 0].set_title('KL Div Loss')

		lossAxs[2, 0].plot(self.__lossDict["total"])
		lossAxs[2, 0].set_title('Total Loss')

		lossFig.tight_layout(pad=1.0)
		if(save):
			lossFig.savefig(f"{save}/lossPlots.png")


def loadModel(filename):
	"""
	Load the scVitalModel object from a file.

	Parameters:
	filename (str): The name of the file to load the model from.

	Returns:
	scVitalModel: The loaded model object.
	"""
	with open(filename, 'rb') as file:
		scvitalModel = pickle.load(file)
	print(f"Model loaded from {filename}")
	return scvitalModel


class EncoderDecoder(nn.Module):
    def __init__(self, encoder, decoder, **kwargs):
        super(EncoderDecoder, self).__init__(**kwargs)
        self.encoder = encoder
        self.decoder = decoder

    def forward(self, encIn, labels, num_classes): 
        latent = self.encoder(encIn)
        #labelsOneHot = F.one_hot(labels, num_classes=num_classes).float()
        labelsOneHot = torch.reshape(F.one_hot(labels.to(torch.int64), num_classes=num_classes).float(),(latent.shape[0], num_classes))
        encOutLabel = torch.cat((latent, labelsOneHot),axis=1)
        decOut = self.decoder(encOutLabel)
        return decOut
    
    def getEncoder(self):
        return self.encoder

    def getDecoder(self):
        return self.decoder
        
    def getGeneIndex(self):
        return self.decoder.getGeneIndexes()

class Encoder(nn.Module):
    def __init__(self, dims, numSpecies):#
        super(Encoder, self).__init__()
        self.hidden0 = nn.Sequential(
            nn.Linear(dims[0] + numSpecies, dims[1], bias=True), # 
            nn.LayerNorm(dims[1]),
            #nn.BatchNorm1d(dims[1]),
            #nn.Dropout(0.05),
            nn.ReLU() #nn.LeakyReLU(0.1)
        )
        self.hidden1 = nn.Sequential(         
            nn.Linear(dims[1], dims[2], bias=True),
            nn.LayerNorm(dims[2]),
            #nn.BatchNorm1d(dims[2]),
            #nn.Dropout(0.05),
            nn.ReLU() #nn.LeakyReLU(0.1)
        )
        self.mu = nn.Sequential(
            nn.Linear(dims[2], dims[3], bias=True)
        )
        self.lnVar = nn.Sequential(
            nn.Linear(dims[2], dims[3], bias=True)
        )
        
        nn.init.kaiming_normal_(self.hidden0[0].weight,nonlinearity='relu')#,a=0.1,nonlinearity='leaky_relu')#
        nn.init.kaiming_normal_(self.hidden1[0].weight,nonlinearity='relu')#,a=0.1,nonlinearity='leaky_relu')#
        self.klDivLoss = 0

    def forward(self, x):
        x = self.hidden0(x)
        x = self.hidden1(x)
        mean = self.mu(x)
        lnVar = self.lnVar(x)
        x = torch.exp(0.5*lnVar) * torch.randn_like(lnVar) + mean
        self.klDivLoss = torch.mean(0.5 * torch.sum(mean**2 + torch.exp(lnVar) - lnVar - 1,dim=1),dim=0)
        return x
    
    
class Decoder(nn.Module):
    def __init__(self, dims, numSpecies, **kwargs):
        self.geneIndexes = kwargs.get("geneIndexes",None)
        self.numSpecies = numSpecies
        super(Decoder, self).__init__()
        self.hidden0 = nn.Sequential(
            nn.Linear(dims[3] + numSpecies, dims[2], bias=True),
            #nn.BatchNorm1d(dims[2]),
            nn.ReLU() #nn.LeakyReLU(0.1)
        )
        self.hidden1 = nn.Sequential(         
            nn.Linear(dims[2], dims[1], bias=True),
            #nn.BatchNorm1d(dims[1]),
            nn.ReLU() #nn.LeakyReLU(0.1)
        )
        self.hidden2 = nn.Sequential(
            nn.Linear(dims[1], dims[0], bias=True)#,dims[0] * 2
            #nn.LeakyReLU(0.1)
        )

        nn.init.kaiming_normal_(self.hidden0[0].weight,nonlinearity='relu')#,a=0.1,nonlinearity='leaky_relu')#
        nn.init.kaiming_normal_(self.hidden1[0].weight,nonlinearity='relu')#,a=0.1,nonlinearity='leaky_relu')#
        #nn.init.kaiming_normal_(self.hidden2[0].weight,nonlinearity='relu')#,a=0.1,nonlinearity='leaky_relu')#

    def forward(self, x):
        x = self.hidden0(x)
        x = self.hidden1(x)
        x = self.hidden2(x)
        return x

    def getGeneIndexes(self):
        return self.geneIndexes
    
class Discriminator(nn.Module):
    def __init__(self, dims, numSpecies):
        super(Discriminator, self).__init__()
        self.hidden0 = nn.Sequential(
            nn.Linear(dims[0], dims[1], bias=True),
            nn.ReLU()
        )
        self.hidden1 = nn.Sequential(
            nn.Linear(dims[1], numSpecies, bias=True)
        )
        #self.out = nn.Sequential(
        #    nn.Identity(numSpecies)
        #)
        
        nn.init.kaiming_normal_(self.hidden0[0].weight,nonlinearity='relu')#,a=0.1,nonlinearity='leaky_relu')#

    def forward(self, x):
        x = self.hidden0(x)
        x = self.hidden1(x)
        #x = self.out(x)
        return x
        



import os
import time
import numpy as np
import scanpy as sc
import pandas as pd
import seaborn as sns
import anndata as an

import matplotlib
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D      
import matplotlib.cm as cm
from matplotlib.patches import Rectangle

import matplotlib.image as img
import matplotlib.patches as patches
import networkx as nx
from networkx.drawing.nx_agraph import graphviz_layout

from sklearn.metrics import roc_curve, roc_auc_score, precision_recall_curve, auc
from scipy.spatial.distance import squareform
from scipy.cluster.hierarchy import linkage
from scipy.cluster import hierarchy 
from scipy.spatial import distance

from collections import defaultdict

def calcPairsLSS(adata, latent, batchName, cellTypeLabel):
    # Check and add simple pairs to the data
    adata = checkPairsAddSimple(adata, batchName, cellTypeLabel)
    # Get unique cell types and their pairs
    allCellTypes, ctPairs = getUniCtPairs(adata, batchName, cellTypeLabel)
    # Calculate LSS metrics
    clustDist, lssAUC, totalDist = calcLSS(adata, latent, cellTypeLabel, batchName, allCellTypes, ctPairs)
    # Return calculated metrics and cell type information
    return clustDist, lssAUC, totalDist, allCellTypes, ctPairs

def calcLSS(adata, latent, cellTypeLabel, batchName, allCellTypes, ctPairs):
    # Calculate cluster distances
    clustDist = calcClustDist(adata, latent, allCellTypes, batchName, cellTypeLabel)
    # Convert cell type pairs to indices
    realPairs = __pairToIndex(np.array(allCellTypes), np.array(allCellTypes), np.array(ctPairs))
    if len(realPairs) == 0:
        print("No pairs or simple pairs") 
        return clustDist, 0, 0
    # Calculate total distance
    totalDist = calcTotalDist(clustDist, realPairs)
    # Calculate AUC for LSS
    lssAUC = calcAUC(clustDist, realPairs)
    # Return calculated distances and AUC
    return clustDist, lssAUC, totalDist


def calcAUC(clustDist, realPairs): 
    # Simplify distance matrix for AUC calculation
    simpDist = np.triu(-1 * clustDist.values) + np.tril(np.full(clustDist.shape[0], fill_value=-2))
    # Initialize true labels matrix
    labelTrue = np.zeros(clustDist.shape)
    # Mark true pairs in the label matrix
    for (i, j) in np.array(realPairs):
        labelTrue[i, j] = 1
    # Flatten the matrices for AUC calculation
    labelPred = simpDist.flatten()
    labelTrueF = np.triu(labelTrue).flatten()
    # Calculate and return AUC of F1 score
    precision, recall, thresholds = precision_recall_curve(labelTrueF, labelPred)
    return(auc(recall, precision))    #auc=roc_auc_score(labelTrue, labelPred)

def calcClustDist(adata, latent, allCellTypes, batchName, cellTypeLabel):
    # Initialize cluster distance matrix
    clustDist = pd.DataFrame(np.zeros((len(allCellTypes), len(allCellTypes))), columns=allCellTypes, index=allCellTypes)
    # Calculate mean latent space values for each cell type and batch
    for hcs in clustDist.index:
        hs, hc = hcs.split("~")
        hMean = np.mean(adata[np.logical_and(adata.obs[cellTypeLabel] == hc, adata.obs[batchName] == hs)].obsm[latent], axis=0)
        for mcs in clustDist.columns:
            ms, mc = mcs.split("~")
            if hcs == mcs:
                clustDist.loc[hcs, mcs] = 0
            else:
                mMean = np.mean(adata[np.logical_and(adata.obs[cellTypeLabel] == mc, adata.obs[batchName] == ms)].obsm[latent], axis=0)
                # Calculate cosine distance between means
                clustDist.loc[hcs, mcs] = np.round(distance.cosine(hMean, mMean), decimals=4)
                # Alternative distance metrics (commented out)
                # clustDist.loc[hcs, mcs] = np.round(mean_squared_error(hMean, mMean), decimals=4)
                # r, _ = scipy.stats.pearsonr(hMean, mMean)
                # clustDist.loc[hcs, mcs] = np.abs(np.round(r, decimals=4) - 1)
    # Return the cluster distance matrix
    return clustDist


def calcTotalDist(clustDist, realPairs): 
    # Initialize total distance
    totalDist = 0
    # Sum distances for all real pairs
    for (i, j) in np.array(realPairs):
        totalDist += clustDist.iloc[i, j]
    # Adjust total distance
    totalDist *= 0.5
    # Return total distance
    return totalDist


def checkPairsAddSimple(adata, batchName, labelName):
    batches = adata.obs[batchName].cat.categories.values
    try:
        adata.uns["pairs"]
    except:
        print("finding pairs")
        adata.uns["pairs"] = simpleCasePairsAdata(adata, batchName, labelName)

    adata.obs["overlapLabel"] = pd.Categorical(__findOverlapLabels(adata.uns["pairs"], adata.obs[labelName]))
    return(adata)

def getUniCtPairs(adata, batchName, labelName):
    batches = adata.obs[batchName].cat.categories.values

    sep="~"
    allCt = []
    for batch in batches:
        cellTypes = np.unique(adata.obs[adata.obs[batchName]==batch][labelName])
        batchCt = [f"{batch}{sep}{ct}" for ct in cellTypes]
        allCt = allCt + batchCt

    pairs = {(ctP[0], ctP[1]) for ctP in adata.uns["pairs"]}

    ctPairs = []
    for p1 in allCt:
        for p2 in allCt:
            name1, ct1 = p1.split(sep)
            name2, ct2 = p2.split(sep)
            if(((ct1==ct2) or ((ct1,ct2) in pairs or (ct2,ct1) in pairs)) and (name1 != name2)):
                ctPairs = ctPairs + [[p1, p2]]

    return(allCt, ctPairs)


def simpleCasePairsAdata(adata, batchName, cellTypeLabel):
    batches = adata.obs[batchName].cat.categories.values
    #for batch in batches:
    #    adata.uns[f"cellTypes{batch}"] = np.unique(adata[adata.obs[batchName]==batch].obs[labelName])

    cellTypesBatch =[set(adata[adata.obs[batchName]==batch].obs[cellTypeLabel]) for batch in batches]

    pairs = []

    for ctB1 in cellTypesBatch:
        for ctB2 in cellTypesBatch:
            for ct in ctB1.intersection(ctB2):
                pairs.append([ct,ct])
            #for ctb1 in ctB1:
            #    for ctb2 in ctB2:
            #        if ctb1 == ctb2:
            #            pairs.append([ctb1,ctb2])
    return(pairs)


#commented with copilot
def __findOverlapLabels(pairs, ogLabels):
    # Initialize an empty list to store overlapping label sets
    simple = []
    # Iterate through each pair of labels
    for pair in pairs:
        setPair = set(pair)
        addPair = True
        # Check if the current pair intersects with any existing label set
        for i, setLab in enumerate(simple):
            if setLab.intersection(setPair):
                # If there's an intersection, merge the sets
                simple[i] = setLab.union(setPair)
                addPair = False
        # If no intersection, add the pair as a new label set
        if addPair:
            simple.append(setPair)
    # Create a dictionary to map labels to simplified labels
    simple = np.unique(simple)
    label2simp = dict()
    for i, setLabels in enumerate(simple):
        label2simp.update(dict(zip(list(setLabels), [f"{i}"] * len(setLabels))))
    # Assign unique simplified labels to any remaining original labels
    totalLabels = len(simple)
    for anno in np.unique(ogLabels):
        if(anno not in label2simp.keys()):
            label2simp.update({anno: f"{totalLabels}"})
            totalLabels += 1
    # Return a list of simplified labels corresponding to the original labels
    simpLabs = np.full(totalLabels, fill_value="" ,dtype=object)
    cellLabels = list(label2simp.keys())
    overlapLabelNum = list(label2simp.values())
    for i in range(totalLabels):
        simpLabs[i] = cellLabels[overlapLabelNum.index(str(i))]
    retSimpleLabels = [simpLabs[int(label2simp[lab])] for lab in ogLabels]
    return(retSimpleLabels)

def __pairToIndex(cellTypes1, cellTypes2, ctPairs):
    overlapCT = []
    for ct1, ct2 in ctPairs:
        if(type(cellTypes1)==np.ndarray):
            try:
                ct1I = np.where(cellTypes1 == ct1)[0][0]
                ct2I = np.where(cellTypes2 == ct2)[0][0]
            except:
                #print("not np array")
                continue
        else:
            ct1I = cellTypes1.index(ct1)
            ct2I = cellTypes2.index(ct2)
        try:
            overlapCT.append([ct1I,ct2I])
        except:
            continue

    return(overlapCT)

def plotHeatLSS(adata, clustDist, latent, allCellTypes, ctPairs, plot=True, save=False):
    condClustDist = squareform(clustDist)
    Z = linkage(condClustDist, 'complete')
    dn = hierarchy.dendrogram(Z, no_plot=True)
    
    mask = np.zeros_like(clustDist)
    mask[np.tril_indices_from(mask)] = True
    mask = mask[:,np.argsort(dn["leaves"])]
    mask = mask[np.argsort(dn["leaves"]),:]
    
    cdMap = sns.clustermap(clustDist, mask=mask, row_linkage=Z, col_linkage=Z, cmap="RdBu",vmin=0,vmax=2)
    cdMap.ax_col_dendrogram.set_visible(False)
    cdMap.ax_heatmap.set_title(latent)
    realPairs = __pairToIndex(np.array(allCellTypes), np.array(allCellTypes), np.array(ctPairs))
    totalDist = 0
    for (i, j) in np.array(realPairs):
        clustered_row_i, clustered_col_j = cdMap.dendrogram_row.reordered_ind.index(i), cdMap.dendrogram_col.reordered_ind.index(j)
        # Draw a rectangle on the clustered heatmap
        if(clustered_row_i < clustered_col_j):
            rect = patches.Rectangle((clustered_col_j , clustered_row_i), 1, 1, linewidth=4, edgecolor='black', facecolor='none')
            cdMap.ax_heatmap.add_patch(rect)
            totalDist += clustDist.iloc[i,j]

    if(not plot):
        plt.close(cdMap.fig)
    if(save): #save
        cdMap.savefig(f"{save}/Clmap_{latent}.svg")

def plotGraphLSS(adata, cellTypeLabel, batchName, clustDist, name="", ctColors=plt.get_cmap('tab10').colors, btColors=None, shapes="ospx^><.....", prog="neato", wLab=False, qCut = 0.28, plot=True, save=False):
    batchDict, annoToColorDict = __getBatchCellDicts(adata, cellTypeLabel, batchName, ctColors, btColors, shapes)

    overlap = np.unique(adata.obs[cellTypeLabel])
    pairs = adata.uns["pairs"]
    batchToColorDict = {lab:batchDict[lab][1] for lab in batchDict.keys()}
    batchToShapeDict = {lab:batchDict[lab][2] for lab in batchDict.keys()}
    
    fig, ax = plt.subplots()
    G = nx.Graph()
    for i in clustDist.columns:
        G.add_node(i)

    allDists = clustDist.to_numpy().flatten()
    cutoff = np.quantile(allDists[allDists>0], qCut)
    
    for i in range(len(clustDist.columns)):
        for j in range(i,len(clustDist.columns)):
            if((i != j) and clustDist.iloc[i,j] < cutoff):
                G.add_edge(clustDist.columns[i], clustDist.index[j], weight=clustDist.iloc[i,j])
    pos = graphviz_layout(G, prog=prog)#, seed=42)
    nx.draw_networkx_edges(G, pos, ax=ax)
    lw=1
    for j,bat in enumerate(list(set([cl.split("~")[0] for cl in clustDist.columns]))):
        nodes = clustDist.columns[[bat==cl.split("~")[0] for cl in clustDist.columns]]        
        ctColors = __getOverColors([label.split("~")[1] for label in nodes], overlap, pairs, annoToColorDict)
        nx.draw_networkx_nodes(G, pos, nodelist=nodes, node_color=ctColors, node_size=150, edgecolors = batchToColorDict[bat],  node_shape=batchToShapeDict[bat], linewidths=lw, ax=ax)#alpha=0.9
    
    fig.suptitle(f"LSS Cut:{qCut}, Prog:{prog}")
    ax.axis("off")
    plt.tight_layout()
    if(not plot):
        plt.close(fig)
    if(save):
        #clustDist.to_csv(f"{save}/multiLSS_{name}scVital.csv")
        #nx.write_adjlist(G, f"{save}/multiLSS_{name}scVital.adjlist")
        #nx.write_weighted_edgelist(G, f"{save}/multiLSS_{name}scVital.weighted.edgelist")
        fig.savefig(f"{save}/graphLSS_{name}.svg", format="svg")
    return batchDict, annoToColorDict

def __group_pairs(pairs):
    # Dictionary to hold the groups
    groups = defaultdict(set)

    # Iterate through each pair
    for name1, name2 in pairs:
        # Find the groups that name1 and name2 belong to
        group1 = next((group for group in groups.values() if name1 in group), None)
        group2 = next((group for group in groups.values() if name2 in group), None)

        if group1 and group2:
            if group1 != group2:
                # Merge the two groups if they are different
                group1.update(group2)
                for name in group2:
                    groups[name] = group1
        elif group1:
            group1.add(name2)
            groups[name2] = group1
        elif group2:
            group2.add(name1)
            groups[name1] = group2
        else:
            # Create a new group if neither name is in any group
            new_group = {name1, name2}
            groups[name1] = new_group
            groups[name2] = new_group

    # Extract unique groups
    unique_groups = set(frozenset(group) for group in groups.values())

    # Convert each group to a list
    return [list(group) for group in unique_groups]

def __findPair(pairs, ctf, visited=None): #modified with copilot
    if visited is None:
        visited = set()
    
    if ctf in visited:
        return None  # or handle the circular reference case appropriately
    
    visited.add(ctf)
    
    check1 = False
    check2 = False
    for ct1,ct2 in pairs:
        if (ctf==ct1):
            if(ct2 in annoToColorDict.keys()):
                return ct2
            else:
                check2 = ct2
        elif (ctf==ct2):
            if(ct1 in annoToColorDict.keys()):
                return ct1
            else:
                check1 = ct1
    if(check1):
        return __findPair(pairs, check1)
    if(check2):
        return __findPair(pairs, check2)

def __getOverColors(ogLabel, overlabel, pairs, colorDict):
    colorOut = ogLabel.copy()
    for i,ctf in enumerate(ogLabel):
        if (ctf not in overlabel):
            ctf = __findPair(pairs, ctf)
        colorOut[i] = colorDict[ctf]
    return colorOut

def __getBatchCellDicts(adata, cellTypeLabel, batchName, ctColors=plt.get_cmap('tab10').colors, btColors=None, shapes="ospx^><....."):
    numBatch=len(adata.obs[batchName].cat.categories)
    if len(shapes) < numBatch:
        shapes = shapes + "."*numBatch

    if len(ctColors) < max(len(adata.obs[cellTypeLabel].cat.categories), len(adata.obs["overlapLabel"].cat.categories)):
        ctColors = list(ctColors) + [ctColors[0]]*(max(len(adata.obs[cellTypeLabel].cat.categories), len(adata.obs["overlapLabel"].cat.categories)) - len(ctColors))

    pairs = adata.uns["pairs"]
    cellTypeDict = dict(zip(adata.obs[cellTypeLabel].cat.categories, list(zip(adata.obs[cellTypeLabel].cat.categories, 
                                                                              ctColors[:len(adata.obs[cellTypeLabel].cat.categories)]))))
    overlapDict = dict(zip(adata.obs["overlapLabel"].cat.categories, list(zip(adata.obs["overlapLabel"].cat.categories, 
                                                                              ctColors[:len(adata.obs["overlapLabel"].cat.categories)]))))
    for overKey in overlapDict.keys():
        cellTypeDict[overKey] = overlapDict[overKey]
    
    gpairs = __group_pairs(pairs)
    for gpair in gpairs:
        overColor = (0,0,0)
        for csc in gpair:
            if csc in overlapDict.keys():
                overColor = overlapDict[csc][1]
                break
        for csc in gpair:
            cellTypeDict[csc] = (csc, overColor)
    
    annoToColorDict = {lab:cellTypeDict[lab][1] for lab in cellTypeDict.keys()}
    
    if(btColors is None):
        btColors = [(i * 0.9/(numBatch-1),i * 0.9/(numBatch-1),i * 0.9/(numBatch-1)) for i in range(numBatch)]
    batchDict = dict(zip(adata.obs[batchName].cat.categories, 
                list(zip(adata.obs[batchName].cat.categories, btColors[:numBatch],shapes[:numBatch]))))
    
    makeLegend(adata.obs[cellTypeLabel], adata.obs[batchName], cellTypeDict, batchDict, name="", save=False)

    return(batchDict, annoToColorDict)


def makeLegend(ctVals, btVals, cellTypeColorDict, batchColorDict, name="", save=False):
    ctColLab = cellTypeColorDict.values()
    btColLab = batchColorDict.values()

    ctLegendEle = [Line2D([0], [0], color=ctc, marker="o", lw=0, label=ctLabel) for ctLabel,ctc in ctColLab]
    spaceLegEle = [Line2D([0], [0], marker='o', lw=0, color='white', markeredgecolor='white', label="")]
    btLegendEle = [Line2D([0], [0], color="white", marker=btShape, lw=0, markeredgecolor=btc, label=btLabel) for i,(btLabel,btc,btShape) in enumerate(btColLab)]

    ctColors = np.array([cellTypeColorDict[ct][1] for ct in ctVals])
    btColors = np.array([batchColorDict[bt][1] for bt in btVals])

    legendEle = btLegendEle + spaceLegEle + ctLegendEle

    fig, ax = plt.subplots()
    plt.legend(handles=legendEle)
    plt.axis('off')
    plt.tight_layout()
    plt.show()
    
    if(save):
        fig.savefig(f"{save}/legend_{name}.svg", format="svg")
    
    #return(legendEle, ctColors, btColors)
